//Copyright (C) 2007 Matthew Johnson

//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.

//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License along
//with this program; if not, write to the Free Software Foundation, Inc.,
//51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
using System;

namespace SVM
{	
	//
	// Kernel Cache
	//
	// l is the number of total data items
	// size is the cache size limit in bytes
	//

	
	//
	// Kernel evaluation
	//
	// the static method k_function is for doing single kernel evaluation
	// the constructor of Kernel prepares to calculate the l*l kernel matrix
	// the member function get_Q is for getting one column from the Q Matrix
	//
	abstract class Kernel
	{
		private Node[][] x;
        private double[] x_square;
		
		private KernelType kernel_type;
		private double degree;
		private double gamma;
		private double coef0;
		
		internal abstract float[] get_Q(int column, int len);
		
		internal virtual void  swap_index(int i, int j)
		{
			do 
			{
				Node[] _ = x[i]; x[i] = x[j]; x[j] = _;
			}
			while (false);
			if (x_square != null)
				do 
				{
					double _ = x_square[i]; x_square[i] = x_square[j]; x_square[j] = _;
				}
				while (false);
		}
		
		private static double tanh(double x)
		{
			double e = System.Math.Exp(x);
			return 1.0 - 2.0 / (e * e + 1);
		}
		
		internal virtual double kernel_function(int i, int j)
		{
			switch (kernel_type)
			{
				
				case KernelType.LINEAR: 
					return dot(x[i], x[j]);

                case KernelType.POLY: 
					return System.Math.Pow(gamma * dot(x[i], x[j]) + coef0, degree);

                case KernelType.RBF: 
					return System.Math.Exp((- gamma) * (x_square[i] + x_square[j] - 2 * dot(x[i], x[j])));

                case KernelType.SIGMOID: 
					return tanh(gamma * dot(x[i], x[j]) + coef0);
				
				default: 
					return 0; // java
				
			}
		}
		
		internal Kernel(int l, Node[][] x_, Parameter param)
		{
			this.kernel_type = param.KernelType;
			this.degree = param.Degree;
			this.gamma = param.Gamma;
			this.coef0 = param.Coefficient0;
			
			x = (Node[][]) x_.Clone();

            if (kernel_type == KernelType.RBF)
			{
				x_square = new double[l];
				for (int i = 0; i < l; i++)
					x_square[i] = dot(x[i], x[i]);
			}
			else
				x_square = null;
		}
		
		internal static double dot(Node[] x, Node[] y)
		{
			double sum = 0;
			int xlen = x.Length;
			int ylen = y.Length;
			int i = 0;
			int j = 0;
			while (i < xlen && j < ylen)
			{
				if (x[i].Index == y[j].Index)
					sum += x[i++].Value * y[j++].Value;
				else
				{
					if (x[i].Index > y[j].Index)
						++j;
					else
						++i;
				}
			}
			return sum;
		}
		
		internal static double k_function(Node[] x, Node[] y, Parameter param)
		{
			switch (param.KernelType)
			{

                case KernelType.LINEAR: 
					return dot(x, y);

                case KernelType.POLY: 
					return System.Math.Pow(param.Gamma * dot(x, y) + param.Coefficient0, param.Degree);

                case KernelType.RBF: 
				{
					double sum = 0;
					int xlen = x.Length;
					int ylen = y.Length;
					int i = 0;
					int j = 0;
					while (i < xlen && j < ylen)
					{
						if (x[i].Index == y[j].Index)
						{
							double d = x[i++].Value - y[j++].Value;
							sum += d * d;
						}
						else if (x[i].Index > y[j].Index)
						{
							sum += y[j].Value * y[j].Value;
							++j;
						}
						else
						{
							sum += x[i].Value * x[i].Value;
							++i;
						}
					}
					
					while (i < xlen)
					{
						sum += x[i].Value * x[i].Value;
						++i;
					}
					
					while (j < ylen)
					{
						sum += y[j].Value * y[j].Value;
						++j;
					}
					
					return System.Math.Exp((- param.Gamma) * sum);
				}

            case KernelType.SIGMOID: 
					return tanh(param.Gamma * dot(x, y) + param.Coefficient0);
				
				default: 
					return 0; // java
				
			}
		}
	}
	
	// Generalized SMO+SVMlight algorithm
	// Solves:
	//
	//	min 0.5(\alpha^T Q \alpha) + b^T \alpha
	//
	//		y^T \alpha = \delta
	//		y_i = +1 or -1
	//		0 <= alpha_i <= Cp for y_i = 1
	//		0 <= alpha_i <= Cn for y_i = -1
	//
	// Given:
	//
	//	Q, b, y, Cp, Cn, and an initial feasible point \alpha
	//	l is the size of vectors and matrices
	//	eps is the stopping criterion
	//
	// solution will be put in \alpha, objective value will be put in obj
	//
	class Solver
	{
		internal int active_size;
		internal sbyte[] y;
		internal double[] G; // gradient of objective function
		internal const sbyte LOWER_BOUND = 0;
		internal const sbyte UPPER_BOUND = 1;
		internal const sbyte FREE = 2;
		internal sbyte[] alpha_status; // LOWER_BOUND, UPPER_BOUND, FREE
		internal double[] alpha;
		internal Kernel Q;
		internal double eps;
		internal double Cp, Cn;
		internal double[] b;
		internal int[] active_set;
		internal double[] G_bar; // gradient, if we treat free variables as 0
		internal int l;
		internal bool unshrinked; // XXX
		
		//UPGRADE_NOTE: Final was removed from the declaration of 'INF '. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1003_3"'
		internal static readonly double INF = System.Double.PositiveInfinity;
		
		internal virtual double get_C(int i)
		{
			return (y[i] > 0)?Cp:Cn;
		}
		internal virtual void  update_alpha_status(int i)
		{
			if (alpha[i] >= get_C(i))
				alpha_status[i] = UPPER_BOUND;
			else if (alpha[i] <= 0)
				alpha_status[i] = LOWER_BOUND;
			else
				alpha_status[i] = FREE;
		}
		internal virtual bool is_upper_bound(int i)
		{
			return alpha_status[i] == UPPER_BOUND;
		}
		internal virtual bool is_lower_bound(int i)
		{
			return alpha_status[i] == LOWER_BOUND;
		}
		internal virtual bool is_free(int i)
		{
			return alpha_status[i] == FREE;
		}
		
		// java: information about solution except alpha,
		// because we cannot return multiple values otherwise...
		internal class SolutionInfo
		{
			internal double obj;
			internal double rho;
			internal double upper_bound_p;
			internal double upper_bound_n;
			internal double r; // for Solver_NU
		}
		
		internal virtual void  swap_index(int i, int j)
		{
			Q.swap_index(i, j);
			do 
			{
				sbyte _ = y[i]; y[i] = y[j]; y[j] = _;
			}
			while (false);
			do 
			{
				double _ = G[i]; G[i] = G[j]; G[j] = _;
			}
			while (false);
			do 
			{
				sbyte _ = alpha_status[i]; alpha_status[i] = alpha_status[j]; alpha_status[j] = _;
			}
			while (false);
			do 
			{
				double _ = alpha[i]; alpha[i] = alpha[j]; alpha[j] = _;
			}
			while (false);
			do 
			{
				double _ = b[i]; b[i] = b[j]; b[j] = _;
			}
			while (false);
			do 
			{
				int _ = active_set[i]; active_set[i] = active_set[j]; active_set[j] = _;
			}
			while (false);
			do 
			{
				double _ = G_bar[i]; G_bar[i] = G_bar[j]; G_bar[j] = _;
			}
			while (false);
		}
		
		internal virtual void  reconstruct_gradient()
		{
			// reconstruct inactive elements of G from G_bar and free variables
			
			if (active_size == l)
				return ;
			
			int i;
			for (i = active_size; i < l; i++)
				G[i] = G_bar[i] + b[i];
			
			for (i = 0; i < active_size; i++)
				if (is_free(i))
				{
					float[] Q_i = Q.get_Q(i, l);
					double alpha_i = alpha[i];
					for (int j = active_size; j < l; j++)
						G[j] += alpha_i * Q_i[j];
				}
		}
		
		internal virtual void  Solve(int l, Kernel Q, double[] b_, sbyte[] y_, double[] alpha_, double Cp, double Cn, double eps, SolutionInfo si, bool shrinking)
		{
			this.l = l;
			this.Q = Q;
			b = new double[b_.Length];
			b_.CopyTo(b, 0);
			y = new sbyte[y_.Length];
			y_.CopyTo(y, 0);
			alpha = new double[alpha_.Length];
			alpha_.CopyTo(alpha, 0);
			this.Cp = Cp;
			this.Cn = Cn;
			this.eps = eps;
			this.unshrinked = false;
			
			// initialize alpha_status
			{
				alpha_status = new sbyte[l];
				for (int i = 0; i < l; i++)
					update_alpha_status(i);
			}
			
			// initialize active set (for shrinking)
			{
				active_set = new int[l];
				for (int i = 0; i < l; i++)
					active_set[i] = i;
				active_size = l;
			}
			
			// initialize gradient
			{
				G = new double[l];
				G_bar = new double[l];
				int i;
				for (i = 0; i < l; i++)
				{
					G[i] = b[i];
					G_bar[i] = 0;
				}
				for (i = 0; i < l; i++)
					if (!is_lower_bound(i))
					{
						float[] Q_i = Q.get_Q(i, l);
						double alpha_i = alpha[i];
						int j;
						for (j = 0; j < l; j++)
							G[j] += alpha_i * Q_i[j];
						if (is_upper_bound(i))
							for (j = 0; j < l; j++)
								G_bar[j] += get_C(i) * Q_i[j];
					}
			}
			
			// optimization step
			
			int iter = 0;
			int counter = System.Math.Min(l, 1000) + 1;
			int[] working_set = new int[2];
			
			while (true)
			{
				// show progress and do shrinking
				
				if (--counter == 0)
				{
					counter = System.Math.Min(l, 1000);
					if (shrinking)
						do_shrinking();
				}
				
				if (select_working_set(working_set) != 0)
				{
					// reconstruct the whole gradient
					reconstruct_gradient();
					// reset active set size and check
					active_size = l;
					if (select_working_set(working_set) != 0)
						break;
					else
						counter = 1; // do shrinking next iteration
				}
				
				int i = working_set[0];
				int j = working_set[1];
				
				++iter;
				
				// update alpha[i] and alpha[j], handle bounds carefully
				
				float[] Q_i = Q.get_Q(i, active_size);
				float[] Q_j = Q.get_Q(j, active_size);
				
				double C_i = get_C(i);
				double C_j = get_C(j);
				
				double old_alpha_i = alpha[i];
				double old_alpha_j = alpha[j];
				
				if (y[i] != y[j])
				{
					double delta = (- G[i] - G[j]) / System.Math.Max(Q_i[i] + Q_j[j] + 2 * Q_i[j], (float) 0);
					double diff = alpha[i] - alpha[j];
					alpha[i] += delta;
					alpha[j] += delta;
					
					if (diff > 0)
					{
						if (alpha[j] < 0)
						{
							alpha[j] = 0;
							alpha[i] = diff;
						}
					}
					else
					{
						if (alpha[i] < 0)
						{
							alpha[i] = 0;
							alpha[j] = - diff;
						}
					}
					if (diff > C_i - C_j)
					{
						if (alpha[i] > C_i)
						{
							alpha[i] = C_i;
							alpha[j] = C_i - diff;
						}
					}
					else
					{
						if (alpha[j] > C_j)
						{
							alpha[j] = C_j;
							alpha[i] = C_j + diff;
						}
					}
				}
				else
				{
					double delta = (G[i] - G[j]) / System.Math.Max(Q_i[i] + Q_j[j] - 2 * Q_i[j], (float) 0);
					double sum = alpha[i] + alpha[j];
					alpha[i] -= delta;
					alpha[j] += delta;
					if (sum > C_i)
					{
						if (alpha[i] > C_i)
						{
							alpha[i] = C_i;
							alpha[j] = sum - C_i;
						}
					}
					else
					{
						if (alpha[j] < 0)
						{
							alpha[j] = 0;
							alpha[i] = sum;
						}
					}
					if (sum > C_j)
					{
						if (alpha[j] > C_j)
						{
							alpha[j] = C_j;
							alpha[i] = sum - C_j;
						}
					}
					else
					{
						if (alpha[i] < 0)
						{
							alpha[i] = 0;
							alpha[j] = sum;
						}
					}
				}
				
				// update G
				
				double delta_alpha_i = alpha[i] - old_alpha_i;
				double delta_alpha_j = alpha[j] - old_alpha_j;
				
				for (int k = 0; k < active_size; k++)
				{
					G[k] += Q_i[k] * delta_alpha_i + Q_j[k] * delta_alpha_j;
				}
				
				// update alpha_status and G_bar
				
				{
					bool ui = is_upper_bound(i);
					bool uj = is_upper_bound(j);
					update_alpha_status(i);
					update_alpha_status(j);
					int k;
					if (ui != is_upper_bound(i))
					{
						Q_i = Q.get_Q(i, l);
						if (ui)
							for (k = 0; k < l; k++)
								G_bar[k] -= C_i * Q_i[k];
						else
							for (k = 0; k < l; k++)
								G_bar[k] += C_i * Q_i[k];
					}
					
					if (uj != is_upper_bound(j))
					{
						Q_j = Q.get_Q(j, l);
						if (uj)
							for (k = 0; k < l; k++)
								G_bar[k] -= C_j * Q_j[k];
						else
							for (k = 0; k < l; k++)
								G_bar[k] += C_j * Q_j[k];
					}
				}
			}
			
			// calculate rho
			
			si.rho = calculate_rho();
			
			// calculate objective value
			{
				double v = 0;
				int i;
				for (i = 0; i < l; i++)
					v += alpha[i] * (G[i] + b[i]);
				
				si.obj = v / 2;
			}
			
			// put back the solution
			{
				for (int i = 0; i < l; i++)
					alpha_[active_set[i]] = alpha[i];
			}
			
			si.upper_bound_p = Cp;
			si.upper_bound_n = Cn;
			
		}
		
		// return 1 if already optimal, return 0 otherwise
		internal virtual int select_working_set(int[] working_set)
		{
			// return i,j which maximize -grad(f)^T d , under constraint
			// if alpha_i == C, d != +1
			// if alpha_i == 0, d != -1
			
			double Gmax1 = - INF; // max { -grad(f)_i * d | y_i*d = +1 }
			int Gmax1_idx = - 1;
			
			double Gmax2 = - INF; // max { -grad(f)_i * d | y_i*d = -1 }
			int Gmax2_idx = - 1;
			
			for (int i = 0; i < active_size; i++)
			{
				if (y[i] == + 1)
				// y = +1
				{
					if (!is_upper_bound(i))
					// d = +1
					{
						if (- G[i] > Gmax1)
						{
							Gmax1 = - G[i];
							Gmax1_idx = i;
						}
					}
					if (!is_lower_bound(i))
					// d = -1
					{
						if (G[i] > Gmax2)
						{
							Gmax2 = G[i];
							Gmax2_idx = i;
						}
					}
				}
				// y = -1
				else
				{
					if (!is_upper_bound(i))
					// d = +1
					{
						if (- G[i] > Gmax2)
						{
							Gmax2 = - G[i];
							Gmax2_idx = i;
						}
					}
					if (!is_lower_bound(i))
					// d = -1
					{
						if (G[i] > Gmax1)
						{
							Gmax1 = G[i];
							Gmax1_idx = i;
						}
					}
				}
			}
			
			if (Gmax1 + Gmax2 < eps)
				return 1;
			
			working_set[0] = Gmax1_idx;
			working_set[1] = Gmax2_idx;
			return 0;
		}
		
		internal virtual void  do_shrinking()
		{
			int i, j, k;
			int[] working_set = new int[2];
			if (select_working_set(working_set) != 0)
				return ;
			i = working_set[0];
			j = working_set[1];
			double Gm1 = (- y[j]) * G[j];
			double Gm2 = y[i] * G[i];
			
			// shrink
			
			for (k = 0; k < active_size; k++)
			{
				if (is_lower_bound(k))
				{
					if (y[k] == + 1)
					{
						if (- G[k] >= Gm1)
							continue;
					}
					else if (- G[k] >= Gm2)
						continue;
				}
				else if (is_upper_bound(k))
				{
					if (y[k] == + 1)
					{
						if (G[k] >= Gm2)
							continue;
					}
					else if (G[k] >= Gm1)
						continue;
				}
				else
					continue;
				
				--active_size;
				swap_index(k, active_size);
				--k; // look at the newcomer
			}
			
			// unshrink, check all variables again before final iterations
			
			if (unshrinked || - (Gm1 + Gm2) > eps * 10)
				return ;
			
			unshrinked = true;
			reconstruct_gradient();
			
			for (k = l - 1; k >= active_size; k--)
			{
				if (is_lower_bound(k))
				{
					if (y[k] == + 1)
					{
						if (- G[k] < Gm1)
							continue;
					}
					else if (- G[k] < Gm2)
						continue;
				}
				else if (is_upper_bound(k))
				{
					if (y[k] == + 1)
					{
						if (G[k] < Gm2)
							continue;
					}
					else if (G[k] < Gm1)
						continue;
				}
				else
					continue;
				
				swap_index(k, active_size);
				active_size++;
				++k; // look at the newcomer
			}
		}
		
		internal virtual double calculate_rho()
		{
			double r;
			int nr_free = 0;
			double ub = INF, lb = - INF, sum_free = 0;
			for (int i = 0; i < active_size; i++)
			{
				double yG = y[i] * G[i];
				
				if (is_lower_bound(i))
				{
					if (y[i] > 0)
						ub = System.Math.Min(ub, yG);
					else
						lb = System.Math.Max(lb, yG);
				}
				else if (is_upper_bound(i))
				{
					if (y[i] < 0)
						ub = System.Math.Min(ub, yG);
					else
						lb = System.Math.Max(lb, yG);
				}
				else
				{
					++nr_free;
					sum_free += yG;
				}
			}
			
			if (nr_free > 0)
				r = sum_free / nr_free;
			else
				r = (ub + lb) / 2;
			
			return r;
		}
	}
	
	//
	// Solver for nu-svm classification and regression
	//
	// additional constraint: e^T \alpha = constant
	//
	sealed class Solver_NU:Solver
	{
		private SolutionInfo si;
		
		internal override void  Solve(int l, Kernel Q, double[] b, sbyte[] y, double[] alpha, double Cp, double Cn, double eps, SolutionInfo si, bool shrinking)
		{
			this.si = si;
			base.Solve(l, Q, b, y, alpha, Cp, Cn, eps, si, shrinking);
		}
		
		internal override int select_working_set(int[] working_set)
		{
			// return i,j which maximize -grad(f)^T d , under constraint
			// if alpha_i == C, d != +1
			// if alpha_i == 0, d != -1
			
			double Gmax1 = - INF; // max { -grad(f)_i * d | y_i = +1, d = +1 }
			int Gmax1_idx = - 1;
			
			double Gmax2 = - INF; // max { -grad(f)_i * d | y_i = +1, d = -1 }
			int Gmax2_idx = - 1;
			
			double Gmax3 = - INF; // max { -grad(f)_i * d | y_i = -1, d = +1 }
			int Gmax3_idx = - 1;
			
			double Gmax4 = - INF; // max { -grad(f)_i * d | y_i = -1, d = -1 }
			int Gmax4_idx = - 1;
			
			for (int i = 0; i < active_size; i++)
			{
				if (y[i] == + 1)
				// y == +1
				{
					if (!is_upper_bound(i))
					// d = +1
					{
						if (- G[i] > Gmax1)
						{
							Gmax1 = - G[i];
							Gmax1_idx = i;
						}
					}
					if (!is_lower_bound(i))
					// d = -1
					{
						if (G[i] > Gmax2)
						{
							Gmax2 = G[i];
							Gmax2_idx = i;
						}
					}
				}
				// y == -1
				else
				{
					if (!is_upper_bound(i))
					// d = +1
					{
						if (- G[i] > Gmax3)
						{
							Gmax3 = - G[i];
							Gmax3_idx = i;
						}
					}
					if (!is_lower_bound(i))
					// d = -1
					{
						if (G[i] > Gmax4)
						{
							Gmax4 = G[i];
							Gmax4_idx = i;
						}
					}
				}
			}
			
			if (System.Math.Max(Gmax1 + Gmax2, Gmax3 + Gmax4) < eps)
				return 1;
			
			if (Gmax1 + Gmax2 > Gmax3 + Gmax4)
			{
				working_set[0] = Gmax1_idx;
				working_set[1] = Gmax2_idx;
			}
			else
			{
				working_set[0] = Gmax3_idx;
				working_set[1] = Gmax4_idx;
			}
			return 0;
		}
		
		internal override void  do_shrinking()
		{
			double Gmax1 = - INF; // max { -grad(f)_i * d | y_i = +1, d = +1 }
			double Gmax2 = - INF; // max { -grad(f)_i * d | y_i = +1, d = -1 }
			double Gmax3 = - INF; // max { -grad(f)_i * d | y_i = -1, d = +1 }
			double Gmax4 = - INF; // max { -grad(f)_i * d | y_i = -1, d = -1 }
			
			int k;
			for (k = 0; k < active_size; k++)
			{
				if (!is_upper_bound(k))
				{
					if (y[k] == + 1)
					{
						if (- G[k] > Gmax1)
							Gmax1 = - G[k];
					}
					else if (- G[k] > Gmax3)
						Gmax3 = - G[k];
				}
				if (!is_lower_bound(k))
				{
					if (y[k] == + 1)
					{
						if (G[k] > Gmax2)
							Gmax2 = G[k];
					}
					else if (G[k] > Gmax4)
						Gmax4 = G[k];
				}
			}
			
			double Gm1 = - Gmax2;
			double Gm2 = - Gmax1;
			double Gm3 = - Gmax4;
			double Gm4 = - Gmax3;
			
			for (k = 0; k < active_size; k++)
			{
				if (is_lower_bound(k))
				{
					if (y[k] == + 1)
					{
						if (- G[k] >= Gm1)
							continue;
					}
					else if (- G[k] >= Gm3)
						continue;
				}
				else if (is_upper_bound(k))
				{
					if (y[k] == + 1)
					{
						if (G[k] >= Gm2)
							continue;
					}
					else if (G[k] >= Gm4)
						continue;
				}
				else
					continue;
				
				--active_size;
				swap_index(k, active_size);
				--k; // look at the newcomer
			}
			
			// unshrink, check all variables again before final iterations
			
			if (unshrinked || System.Math.Max(- (Gm1 + Gm2), - (Gm3 + Gm4)) > eps * 10)
				return ;
			
			unshrinked = true;
			reconstruct_gradient();
			
			for (k = l - 1; k >= active_size; k--)
			{
				if (is_lower_bound(k))
				{
					if (y[k] == + 1)
					{
						if (- G[k] < Gm1)
							continue;
					}
					else if (- G[k] < Gm3)
						continue;
				}
				else if (is_upper_bound(k))
				{
					if (y[k] == + 1)
					{
						if (G[k] < Gm2)
							continue;
					}
					else if (G[k] < Gm4)
						continue;
				}
				else
					continue;
				
				swap_index(k, active_size);
				active_size++;
				++k; // look at the newcomer
			}
		}
		
		internal override double calculate_rho()
		{
			int nr_free1 = 0, nr_free2 = 0;
			double ub1 = INF, ub2 = INF;
			double lb1 = - INF, lb2 = - INF;
			double sum_free1 = 0, sum_free2 = 0;
			
			for (int i = 0; i < active_size; i++)
			{
				if (y[i] == + 1)
				{
					if (is_lower_bound(i))
						ub1 = System.Math.Min(ub1, G[i]);
					else if (is_upper_bound(i))
						lb1 = System.Math.Max(lb1, G[i]);
					else
					{
						++nr_free1;
						sum_free1 += G[i];
					}
				}
				else
				{
					if (is_lower_bound(i))
						ub2 = System.Math.Min(ub2, G[i]);
					else if (is_upper_bound(i))
						lb2 = System.Math.Max(lb2, G[i]);
					else
					{
						++nr_free2;
						sum_free2 += G[i];
					}
				}
			}
			
			double r1, r2;
			if (nr_free1 > 0)
				r1 = sum_free1 / nr_free1;
			else
				r1 = (ub1 + lb1) / 2;
			
			if (nr_free2 > 0)
				r2 = sum_free2 / nr_free2;
			else
				r2 = (ub2 + lb2) / 2;
			
			si.r = (r1 + r2) / 2;
			return (r1 - r2) / 2;
		}
	}
	
	//
	// Q matrices for various formulations
	//
	class SVC_Q:Kernel
	{
		//UPGRADE_NOTE: Final was removed from the declaration of 'y '. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1003_3"'
		private sbyte[] y;
		//UPGRADE_NOTE: Final was removed from the declaration of 'cache '. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1003_3"'
		private Cache cache;
		
		internal SVC_Q(Problem prob, Parameter param, sbyte[] y_):base(prob.Count, prob.X, param)
		{
			y = new sbyte[y_.Length];
			y_.CopyTo(y, 0);
			//UPGRADE_WARNING: Data types in Visual C# might be different.  Verify the accuracy of narrowing conversions. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1042_3"'
			cache = new Cache(prob.Count, (int) (param.CacheSize * (1 << 20)));
		}
		
		internal override float[] get_Q(int i, int len)
		{
			float[][] data = new float[1][];
			int start;
			if ((start = cache.get_data(i, data, len)) < len)
			{
				for (int j = start; j < len; j++)
				{
					//UPGRADE_WARNING: Data types in Visual C# might be different.  Verify the accuracy of narrowing conversions. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1042_3"'
					data[0][j] = (float) (y[i] * y[j] * kernel_function(i, j));
				}
			}
			return data[0];
		}
		
		internal override void  swap_index(int i, int j)
		{
			cache.swap_index(i, j);
			base.swap_index(i, j);
			do 
			{
				sbyte _ = y[i]; y[i] = y[j]; y[j] = _;
			}
			while (false);
		}
	}
	
	class ONE_CLASS_Q:Kernel
	{
		//UPGRADE_NOTE: Final was removed from the declaration of 'cache '. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1003_3"'
		private Cache cache;
		
		internal ONE_CLASS_Q(Problem prob, Parameter param):base(prob.Count, prob.X, param)
		{
			//UPGRADE_WARNING: Data types in Visual C# might be different.  Verify the accuracy of narrowing conversions. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1042_3"'
			cache = new Cache(prob.Count, (int) (param.CacheSize * (1 << 20)));
		}
		
		internal override float[] get_Q(int i, int len)
		{
			float[][] data = new float[1][];
			int start;
			if ((start = cache.get_data(i, data, len)) < len)
			{
				for (int j = start; j < len; j++)
				{
					//UPGRADE_WARNING: Data types in Visual C# might be different.  Verify the accuracy of narrowing conversions. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1042_3"'
					data[0][j] = (float) kernel_function(i, j);
				}
			}
			return data[0];
		}
		
		internal override void  swap_index(int i, int j)
		{
			cache.swap_index(i, j);
			base.swap_index(i, j);
		}
	}
	
	class SVR_Q:Kernel
	{
		//UPGRADE_NOTE: Final was removed from the declaration of 'l '. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1003_3"'
		private int l;
		//UPGRADE_NOTE: Final was removed from the declaration of 'cache '. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1003_3"'
		private Cache cache;
		//UPGRADE_NOTE: Final was removed from the declaration of 'sign '. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1003_3"'
		private sbyte[] sign;
		//UPGRADE_NOTE: Final was removed from the declaration of 'index '. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1003_3"'
		private int[] index;
		private int next_buffer;
		private float[][] buffer;
		
		internal SVR_Q(Problem prob, Parameter param):base(prob.Count, prob.X, param)
		{
			l = prob.Count;
			//UPGRADE_WARNING: Data types in Visual C# might be different.  Verify the accuracy of narrowing conversions. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1042_3"'
			cache = new Cache(l, (int) (param.CacheSize * (1 << 20)));
			sign = new sbyte[2 * l];
			index = new int[2 * l];
			for (int k = 0; k < l; k++)
			{
				sign[k] = 1;
				sign[k + l] = - 1;
				index[k] = k;
				index[k + l] = k;
			}
			buffer = new float[2][];
			for (int i = 0; i < 2; i++)
			{
				buffer[i] = new float[2 * l];
			}
			next_buffer = 0;
		}
		
		internal override void  swap_index(int i, int j)
		{
			do 
			{
				sbyte _ = sign[i]; sign[i] = sign[j]; sign[j] = _;
			}
			while (false);
			do 
			{
				int _ = index[i]; index[i] = index[j]; index[j] = _;
			}
			while (false);
		}
		
		internal override float[] get_Q(int i, int len)
		{
			float[][] data = new float[1][];
			int real_i = index[i];
			if (cache.get_data(real_i, data, l) < l)
			{
				for (int j = 0; j < l; j++)
				{
					//UPGRADE_WARNING: Data types in Visual C# might be different.  Verify the accuracy of narrowing conversions. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1042_3"'
					data[0][j] = (float) kernel_function(real_i, j);
				}
			}
			
			// reorder and copy
			float[] buf = buffer[next_buffer];
			next_buffer = 1 - next_buffer;
			sbyte si = sign[i];
			for (int j = 0; j < len; j++)
				buf[j] = si * sign[j] * data[0][index[j]];
			return buf;
		}
	}
	
	internal static class Procedures
	{
        private static Random RAND = new Random();
		//
		// construct and solve various formulations
		//
		private static void  solve_c_svc(Problem prob, Parameter param, double[] alpha, Solver.SolutionInfo si, double Cp, double Cn)
		{
			int l = prob.Count;
			double[] minus_ones = new double[l];
			sbyte[] y = new sbyte[l];
			
			int i;
			
			for (i = 0; i < l; i++)
			{
				alpha[i] = 0;
				minus_ones[i] = - 1;
				if (prob.Y[i] > 0)
					y[i] = (sbyte) (+ 1);
				else
					y[i] = - 1;
			}
			
			Solver s = new Solver();
			s.Solve(l, new SVC_Q(prob, param, y), minus_ones, y, alpha, Cp, Cn, param.EPS, si, param.Shrinking);
			
			double sum_alpha = 0;
			for (i = 0; i < l; i++)
				sum_alpha += alpha[i];
					
			for (i = 0; i < l; i++)
				alpha[i] *= y[i];
		}
		
		private static void  solve_nu_svc(Problem prob, Parameter param, double[] alpha, Solver.SolutionInfo si)
		{
			int i;
			int l = prob.Count;
			double nu = param.Nu;
			
			sbyte[] y = new sbyte[l];
			
			for (i = 0; i < l; i++)
				if (prob.Y[i] > 0)
					y[i] = (sbyte) (+ 1);
				else
					y[i] = - 1;
			
			double sum_pos = nu * l / 2;
			double sum_neg = nu * l / 2;
			
			for (i = 0; i < l; i++)
				if (y[i] == + 1)
				{
					alpha[i] = System.Math.Min(1.0, sum_pos);
					sum_pos -= alpha[i];
				}
				else
				{
					alpha[i] = System.Math.Min(1.0, sum_neg);
					sum_neg -= alpha[i];
				}
			
			double[] zeros = new double[l];
			
			for (i = 0; i < l; i++)
				zeros[i] = 0;
			
			Solver_NU s = new Solver_NU();
			s.Solve(l, new SVC_Q(prob, param, y), zeros, y, alpha, 1.0, 1.0, param.EPS, si, param.Shrinking);
			double r = si.r;
						
			for (i = 0; i < l; i++)
				alpha[i] *= y[i] / r;
			
			si.rho /= r;
			si.obj /= (r * r);
			si.upper_bound_p = 1 / r;
			si.upper_bound_n = 1 / r;
		}
		
		private static void  solve_one_class(Problem prob, Parameter param, double[] alpha, Solver.SolutionInfo si)
		{
			int l = prob.Count;
			double[] zeros = new double[l];
			sbyte[] ones = new sbyte[l];
			int i;
			
			//UPGRADE_WARNING: Data types in Visual C# might be different.  Verify the accuracy of narrowing conversions. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1042_3"'
			int n = (int) (param.Nu * prob.Count); // # of alpha's at upper bound
			
			for (i = 0; i < n; i++)
				alpha[i] = 1;
			alpha[n] = param.Nu * prob.Count - n;
			for (i = n + 1; i < l; i++)
				alpha[i] = 0;
			
			for (i = 0; i < l; i++)
			{
				zeros[i] = 0;
				ones[i] = 1;
			}
			
			Solver s = new Solver();
			s.Solve(l, new ONE_CLASS_Q(prob, param), zeros, ones, alpha, 1.0, 1.0, param.EPS, si, param.Shrinking);
		}
		
		private static void  solve_epsilon_svr(Problem prob, Parameter param, double[] alpha, Solver.SolutionInfo si)
		{
			int l = prob.Count;
			double[] alpha2 = new double[2 * l];
			double[] linear_term = new double[2 * l];
			sbyte[] y = new sbyte[2 * l];
			int i;
			
			for (i = 0; i < l; i++)
			{
				alpha2[i] = 0;
				linear_term[i] = param.P - prob.Y[i];
				y[i] = 1;
				
				alpha2[i + l] = 0;
				linear_term[i + l] = param.P + prob.Y[i];
				y[i + l] = - 1;
			}
			
			Solver s = new Solver();
			s.Solve(2 * l, new SVR_Q(prob, param), linear_term, y, alpha2, param.C, param.C, param.EPS, si, param.Shrinking);
			
			double sum_alpha = 0;
			for (i = 0; i < l; i++)
			{
				alpha[i] = alpha2[i] - alpha2[i + l];
				sum_alpha += System.Math.Abs(alpha[i]);
			}
		}
		
		private static void  solve_nu_svr(Problem prob, Parameter param, double[] alpha, Solver.SolutionInfo si)
		{
			int l = prob.Count;
			double C = param.C;
			double[] alpha2 = new double[2 * l];
			double[] linear_term = new double[2 * l];
			sbyte[] y = new sbyte[2 * l];
			int i;
			
			double sum = C * param.Nu * l / 2;
			for (i = 0; i < l; i++)
			{
				alpha2[i] = alpha2[i + l] = System.Math.Min(sum, C);
				sum -= alpha2[i];
				
				linear_term[i] = - prob.Y[i];
				y[i] = 1;
				
				linear_term[i + l] = prob.Y[i];
				y[i + l] = - 1;
			}
			
			Solver_NU s = new Solver_NU();
			s.Solve(2 * l, new SVR_Q(prob, param), linear_term, y, alpha2, C, C, param.EPS, si, param.Shrinking);
						
			for (i = 0; i < l; i++)
				alpha[i] = alpha2[i] - alpha2[i + l];
		}
		
		//
		// decision_function
		//
		internal class decision_function
		{
			internal double[] alpha;
			internal double rho;
		}
		
		
		internal static decision_function svm_train_one(Problem prob, Parameter param, double Cp, double Cn)
		{
			double[] alpha = new double[prob.Count];
			Solver.SolutionInfo si = new Solver.SolutionInfo();
			switch (param.SvmType)
			{
				
				case SvmType.C_SVC: 
					solve_c_svc(prob, param, alpha, si, Cp, Cn);
					break;

                case SvmType.NU_SVC: 
					solve_nu_svc(prob, param, alpha, si);
					break;

                case SvmType.ONE_CLASS: 
					solve_one_class(prob, param, alpha, si);
					break;

                case SvmType.EPSILON_SVR: 
					solve_epsilon_svr(prob, param, alpha, si);
					break;

                case SvmType.NU_SVR: 
					solve_nu_svr(prob, param, alpha, si);
					break;
				}
			
			// output SVs
			
			int nSV = 0;
			int nBSV = 0;
			for (int i = 0; i < prob.Count; i++)
			{
				if (System.Math.Abs(alpha[i]) > 0)
				{
					++nSV;
					if (prob.Y[i] > 0)
					{
						if (System.Math.Abs(alpha[i]) >= si.upper_bound_p)
							++nBSV;
					}
					else
					{
						if (System.Math.Abs(alpha[i]) >= si.upper_bound_n)
							++nBSV;
					}
				}
			}
			
			decision_function f = new decision_function();
			f.alpha = alpha;
			f.rho = si.rho;
			return f;
		}
		
		// Platt's binary SVM Probablistic Output: an improvement from Lin et al.
		private static void  sigmoid_train(int l, double[] dec_values, double[] labels, double[] probAB)
		{
			double A, B;
			double prior1 = 0, prior0 = 0;
			int i;
			
			for (i = 0; i < l; i++)
				if (labels[i] > 0)
					prior1 += 1;
				else
					prior0 += 1;
			
			int max_iter = 100; // Maximal number of iterations
			double min_step = 1e-10; // Minimal step taken in line search
			double sigma = 1e-3; // For numerically strict PD of Hessian
			double eps = 1e-5;
			double hiTarget = (prior1 + 1.0) / (prior1 + 2.0);
			double loTarget = 1 / (prior0 + 2.0);
			double[] t = new double[l];
			double fApB, p, q, h11, h22, h21, g1, g2, det, dA, dB, gd, stepsize;
			double newA, newB, newf, d1, d2;
			int iter;
			
			// Initial Point and Initial Fun Value
			A = 0.0; B = System.Math.Log((prior0 + 1.0) / (prior1 + 1.0));
			double fval = 0.0;
			
			for (i = 0; i < l; i++)
			{
				if (labels[i] > 0)
					t[i] = hiTarget;
				else
					t[i] = loTarget;
				fApB = dec_values[i] * A + B;
				if (fApB >= 0)
					fval += t[i] * fApB + System.Math.Log(1 + System.Math.Exp(- fApB));
				else
					fval += (t[i] - 1) * fApB + System.Math.Log(1 + System.Math.Exp(fApB));
			}
			for (iter = 0; iter < max_iter; iter++)
			{
				// Update Gradient and Hessian (use H' = H + sigma I)
				h11 = sigma; // numerically ensures strict PD
				h22 = sigma;
				h21 = 0.0; g1 = 0.0; g2 = 0.0;
				for (i = 0; i < l; i++)
				{
					fApB = dec_values[i] * A + B;
					if (fApB >= 0)
					{
						p = System.Math.Exp(- fApB) / (1.0 + System.Math.Exp(- fApB));
						q = 1.0 / (1.0 + System.Math.Exp(- fApB));
					}
					else
					{
						p = 1.0 / (1.0 + System.Math.Exp(fApB));
						q = System.Math.Exp(fApB) / (1.0 + System.Math.Exp(fApB));
					}
					d2 = p * q;
					h11 += dec_values[i] * dec_values[i] * d2;
					h22 += d2;
					h21 += dec_values[i] * d2;
					d1 = t[i] - p;
					g1 += dec_values[i] * d1;
					g2 += d1;
				}
				
				// Stopping Criteria
				if (System.Math.Abs(g1) < eps && System.Math.Abs(g2) < eps)
					break;
				
				// Finding Newton direction: -inv(H') * g
				det = h11 * h22 - h21 * h21;
				dA = (- (h22 * g1 - h21 * g2)) / det;
				dB = (- ((- h21) * g1 + h11 * g2)) / det;
				gd = g1 * dA + g2 * dB;
				
				
				stepsize = 1; // Line Search
				while (stepsize >= min_step)
				{
					newA = A + stepsize * dA;
					newB = B + stepsize * dB;
					
					// New function value
					newf = 0.0;
					for (i = 0; i < l; i++)
					{
						fApB = dec_values[i] * newA + newB;
						if (fApB >= 0)
							newf += t[i] * fApB + System.Math.Log(1 + System.Math.Exp(- fApB));
						else
							newf += (t[i] - 1) * fApB + System.Math.Log(1 + System.Math.Exp(fApB));
					}
					// Check sufficient decrease
					if (newf < fval + 0.0001 * stepsize * gd)
					{
						A = newA; B = newB; fval = newf;
						break;
					}
					else
						stepsize = stepsize / 2.0;
				}
				
				if (stepsize < min_step)
				{
					System.Console.Error.Write("Line search fails in two-class probability estimates\n");
					break;
				}
			}
			
			if (iter >= max_iter)
				System.Console.Error.Write("Reaching maximal iterations in two-class probability estimates\n");
			probAB[0] = A; probAB[1] = B;
		}
		
		private static double sigmoid_predict(double decision_value, double A, double B)
		{
			double fApB = decision_value * A + B;
			if (fApB >= 0)
				return System.Math.Exp(- fApB) / (1.0 + System.Math.Exp(- fApB));
			else
				return 1.0 / (1 + System.Math.Exp(fApB));
		}
		
		// Method 2 from the multiclass_prob paper by Wu, Lin, and Weng
		private static void  multiclass_probability(int k, double[][] r, double[] p)
		{
			int t;
			int iter = 0, max_iter = 100;
			double[][] Q = new double[k][];
			for (int i = 0; i < k; i++)
			{
				Q[i] = new double[k];
			}
			double[] Qp = new double[k];
			double pQp, eps = 0.001;
			
			for (t = 0; t < k; t++)
			{
				p[t] = 1.0 / k; // Valid if k = 1
				Q[t][t] = 0;
				for (int j = 0; j < t; j++)
				{
					Q[t][t] += r[j][t] * r[j][t];
					Q[t][j] = Q[j][t];
				}
				for (int j = t + 1; j < k; j++)
				{
					Q[t][t] += r[j][t] * r[j][t];
					Q[t][j] = (- r[j][t]) * r[t][j];
				}
			}
			for (iter = 0; iter < max_iter; iter++)
			{
				// stopping condition, recalculate QP,pQP for numerical accuracy
				pQp = 0;
				for (t = 0; t < k; t++)
				{
					Qp[t] = 0;
					for (int j = 0; j < k; j++)
						Qp[t] += Q[t][j] * p[j];
					pQp += p[t] * Qp[t];
				}
				double max_error = 0;
				for (t = 0; t < k; t++)
				{
					double error = System.Math.Abs(Qp[t] - pQp);
					if (error > max_error)
						max_error = error;
				}
				if (max_error < eps)
					break;
				
				for (t = 0; t < k; t++)
				{
					double diff = (- Qp[t] + pQp) / Q[t][t];
					p[t] += diff;
					pQp = (pQp + diff * (diff * Q[t][t] + 2 * Qp[t])) / (1 + diff) / (1 + diff);
					for (int j = 0; j < k; j++)
					{
						Qp[j] = (Qp[j] + diff * Q[t][j]) / (1 + diff);
						p[j] /= (1 + diff);
					}
				}
			}
			if (iter >= max_iter)
				System.Console.Error.Write("Exceeds max_iter in multiclass_prob\n");
		}
		
		// Cross-validation decision values for probability estimates
		private static void  svm_binary_svc_probability(Problem prob, Parameter param, double Cp, double Cn, double[] probAB)
		{
			int i;
			int nr_fold = 5;
			int[] perm = new int[prob.Count];
			double[] dec_values = new double[prob.Count];
			
			// random shuffle
			for (i = 0; i < prob.Count; i++)
				perm[i] = i;
			for (i = 0; i < prob.Count; i++)
			{
				//UPGRADE_WARNING: Data types in Visual C# might be different.  Verify the accuracy of narrowing conversions. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1042_3"'
				int j = i + (int) (RAND.NextDouble() * (prob.Count - i));
				do 
				{
					int _ = perm[i]; perm[i] = perm[j]; perm[j] = _;
				}
				while (false);
			}
			for (i = 0; i < nr_fold; i++)
			{
				int begin = i * prob.Count / nr_fold;
				int end = (i + 1) * prob.Count / nr_fold;
				int j, k;
				Problem subprob = new Problem();
				
				subprob.Count = prob.Count - (end - begin);
				subprob.X = new Node[subprob.Count][];
				subprob.Y = new double[subprob.Count];
				
				k = 0;
				for (j = 0; j < begin; j++)
				{
					subprob.X[k] = prob.X[perm[j]];
					subprob.Y[k] = prob.Y[perm[j]];
					++k;
				}
				for (j = end; j < prob.Count; j++)
				{
					subprob.X[k] = prob.X[perm[j]];
					subprob.Y[k] = prob.Y[perm[j]];
					++k;
				}
				int p_count = 0, n_count = 0;
				for (j = 0; j < k; j++)
					if (subprob.Y[j] > 0)
						p_count++;
					else
						n_count++;
				
				if (p_count == 0 && n_count == 0)
					for (j = begin; j < end; j++)
						dec_values[perm[j]] = 0;
				else if (p_count > 0 && n_count == 0)
					for (j = begin; j < end; j++)
						dec_values[perm[j]] = 1;
				else if (p_count == 0 && n_count > 0)
					for (j = begin; j < end; j++)
						dec_values[perm[j]] = - 1;
				else
				{
					Parameter subparam = (Parameter) param.Clone();
					subparam.Probability = false;
					subparam.C = 1.0;
					subparam.WeightCount = 2;
					subparam.WeightLabels = new int[2];
					subparam.Weights = new double[2];
					subparam.WeightLabels[0] = + 1;
					subparam.WeightLabels[1] = - 1;
					subparam.Weights[0] = Cp;
					subparam.Weights[1] = Cn;
					Model submodel = svm_train(subprob, subparam);
					for (j = begin; j < end; j++)
					{
						double[] dec_value = new double[1];
						svm_predict_values(submodel, prob.X[perm[j]], dec_value);
						dec_values[perm[j]] = dec_value[0];
						// ensure +1 -1 order; reason not using CV subroutine
						dec_values[perm[j]] *= submodel.ClassLabels[0];
					}
				}
			}
			sigmoid_train(prob.Count, dec_values, prob.Y, probAB);
		}
		
		// Return parameter of a Laplace distribution 
		private static double svm_svr_probability(Problem prob, Parameter param)
		{
			int i;
			int nr_fold = 5;
			double[] ymv = new double[prob.Count];
			double mae = 0;
			
			Parameter newparam = (Parameter) param.Clone();
			newparam.Probability = false;
			svm_cross_validation(prob, newparam, nr_fold, ymv);
			for (i = 0; i < prob.Count; i++)
			{
				ymv[i] = prob.Y[i] - ymv[i];
				mae += System.Math.Abs(ymv[i]);
			}
			mae /= prob.Count;
			double std = System.Math.Sqrt(2 * mae * mae);
			int count = 0;
			mae = 0;
			for (i = 0; i < prob.Count; i++)
				if (System.Math.Abs(ymv[i]) > 5 * std)
					count = count + 1;
				else
					mae += System.Math.Abs(ymv[i]);
			mae /= (prob.Count - count);
			System.Console.Error.Write("Prob. model for test data: target value = predicted value + z,\nz: Laplace distribution e^(-|z|/sigma)/(2sigma),sigma=" + mae + "\n");
			return mae;
		}
		
		//
		// Interface functions
		//
		public static Model svm_train(Problem prob, Parameter param)
		{
			Model model = new Model();
			model.Parameter = param;

            if (param.SvmType == SvmType.ONE_CLASS || param.SvmType == SvmType.EPSILON_SVR || param.SvmType == SvmType.NU_SVR)
			{
				// regression or one-class-svm
				model.NumberOfClasses = 2;
				model.ClassLabels = null;
				model.NumberOfSVPerClass = null;
				model.PairwiseProbabilityA = null; model.PairwiseProbabilityB = null;
				model.SupportVectorCoefficients = new double[1][];

                if (param.Probability && (param.SvmType == SvmType.EPSILON_SVR || param.SvmType == SvmType.NU_SVR))
				{
					model.PairwiseProbabilityA = new double[1];
					model.PairwiseProbabilityA[0] = svm_svr_probability(prob, param);
				}
				
				decision_function f = svm_train_one(prob, param, 0, 0);
				model.Rho = new double[1];
				model.Rho[0] = f.rho;
				
				int nSV = 0;
				int i;
				for (i = 0; i < prob.Count; i++)
					if (System.Math.Abs(f.alpha[i]) > 0)
						++nSV;
				model.SupportVectorCount = nSV;
				model.SupportVectors = new Node[nSV][];
				model.SupportVectorCoefficients[0] = new double[nSV];
				int j = 0;
				for (i = 0; i < prob.Count; i++)
					if (System.Math.Abs(f.alpha[i]) > 0)
					{
						model.SupportVectors[j] = prob.X[i];
						model.SupportVectorCoefficients[0][j] = f.alpha[i];
						++j;
					}
			}
			else
			{
				// classification
				// find out the number of classes
				int l = prob.Count;
				int max_nr_class = 16;
				int nr_class = 0;
				int[] label = new int[max_nr_class];
				int[] count = new int[max_nr_class];
				int[] index = new int[l];
				
				int i;
				for (i = 0; i < l; i++)
				{
					//UPGRADE_WARNING: Data types in Visual C# might be different.  Verify the accuracy of narrowing conversions. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1042_3"'
					int this_label = (int) prob.Y[i];
					int j;
					for (j = 0; j < nr_class; j++)
						if (this_label == label[j])
						{
							++count[j];
							break;
						}
					index[i] = j;
					if (j == nr_class)
					{
						if (nr_class == max_nr_class)
						{
							max_nr_class *= 2;
							int[] new_data = new int[max_nr_class];
							Array.Copy(label, 0, new_data, 0, label.Length);
							label = new_data;
							
							new_data = new int[max_nr_class];
							Array.Copy(count, 0, new_data, 0, count.Length);
							count = new_data;
						}
						label[nr_class] = this_label;
						count[nr_class] = 1;
						++nr_class;
					}
				}
				
				// group training data of the same class
				
				int[] start = new int[nr_class];
				start[0] = 0;
				for (i = 1; i < nr_class; i++)
					start[i] = start[i - 1] + count[i - 1];
				
				Node[][] x = new Node[l][];
				
				for (i = 0; i < l; i++)
				{
					x[start[index[i]]] = prob.X[i];
					++start[index[i]];
				}
				
				start[0] = 0;
				for (i = 1; i < nr_class; i++)
					start[i] = start[i - 1] + count[i - 1];
				
				// calculate weighted C
				
				double[] weighted_C = new double[nr_class];
				for (i = 0; i < nr_class; i++)
					weighted_C[i] = param.C;
				for (i = 0; i < param.WeightCount; i++)
				{
					int j;
					for (j = 0; j < nr_class; j++)
						if (param.WeightLabels[i] == label[j])
							break;
					if (j == nr_class)
						System.Console.Error.Write("warning: class label " + param.WeightLabels[i] + " specified in weight is not found\n");
					else
						weighted_C[j] *= param.Weights[i];
				}
				
				// train k*(k-1)/2 models
				
				bool[] nonzero = new bool[l];
				for (i = 0; i < l; i++)
					nonzero[i] = false;
				decision_function[] f = new decision_function[nr_class * (nr_class - 1) / 2];
				
				double[] probA = null, probB = null;
				if (param.Probability)
				{
					probA = new double[nr_class * (nr_class - 1) / 2];
					probB = new double[nr_class * (nr_class - 1) / 2];
				}
				
				int p = 0;
				for (i = 0; i < nr_class; i++)
					for (int j = i + 1; j < nr_class; j++)
					{
						Problem sub_prob = new Problem();
						int si = start[i], sj = start[j];
						int ci = count[i], cj = count[j];
						sub_prob.Count = ci + cj;
						sub_prob.X = new Node[sub_prob.Count][];
						sub_prob.Y = new double[sub_prob.Count];
						int k;
						for (k = 0; k < ci; k++)
						{
							sub_prob.X[k] = x[si + k];
							sub_prob.Y[k] = + 1;
						}
						for (k = 0; k < cj; k++)
						{
							sub_prob.X[ci + k] = x[sj + k];
							sub_prob.Y[ci + k] = - 1;
						}
						
						if (param.Probability)
						{
							double[] probAB = new double[2];
							svm_binary_svc_probability(sub_prob, param, weighted_C[i], weighted_C[j], probAB);
							probA[p] = probAB[0];
							probB[p] = probAB[1];
						}
						
						f[p] = svm_train_one(sub_prob, param, weighted_C[i], weighted_C[j]);
						for (k = 0; k < ci; k++)
							if (!nonzero[si + k] && System.Math.Abs(f[p].alpha[k]) > 0)
								nonzero[si + k] = true;
						for (k = 0; k < cj; k++)
							if (!nonzero[sj + k] && System.Math.Abs(f[p].alpha[ci + k]) > 0)
								nonzero[sj + k] = true;
						++p;
					}
				
				// build output
				
				model.NumberOfClasses = nr_class;
				
				model.ClassLabels = new int[nr_class];
				for (i = 0; i < nr_class; i++)
					model.ClassLabels[i] = label[i];
				
				model.Rho = new double[nr_class * (nr_class - 1) / 2];
				for (i = 0; i < nr_class * (nr_class - 1) / 2; i++)
					model.Rho[i] = f[i].rho;
				
				if (param.Probability)
				{
					model.PairwiseProbabilityA = new double[nr_class * (nr_class - 1) / 2];
					model.PairwiseProbabilityB = new double[nr_class * (nr_class - 1) / 2];
					for (i = 0; i < nr_class * (nr_class - 1) / 2; i++)
					{
						model.PairwiseProbabilityA[i] = probA[i];
						model.PairwiseProbabilityB[i] = probB[i];
					}
				}
				else
				{
					model.PairwiseProbabilityA = null;
					model.PairwiseProbabilityB = null;
				}
				
				int nnz = 0;
				int[] nz_count = new int[nr_class];
				model.NumberOfSVPerClass = new int[nr_class];
				for (i = 0; i < nr_class; i++)
				{
					int nSV = 0;
					for (int j = 0; j < count[i]; j++)
						if (nonzero[start[i] + j])
						{
							++nSV;
							++nnz;
						}
					model.NumberOfSVPerClass[i] = nSV;
					nz_count[i] = nSV;
				}
				
				model.SupportVectorCount = nnz;
				model.SupportVectors = new Node[nnz][];
				p = 0;
				for (i = 0; i < l; i++)
					if (nonzero[i])
						model.SupportVectors[p++] = x[i];
				
				int[] nz_start = new int[nr_class];
				nz_start[0] = 0;
				for (i = 1; i < nr_class; i++)
					nz_start[i] = nz_start[i - 1] + nz_count[i - 1];
				
				model.SupportVectorCoefficients = new double[nr_class - 1][];
				for (i = 0; i < nr_class - 1; i++)
					model.SupportVectorCoefficients[i] = new double[nnz];
				
				p = 0;
				for (i = 0; i < nr_class; i++)
					for (int j = i + 1; j < nr_class; j++)
					{
						// classifier (i,j): coefficients with
						// i are in sv_coef[j-1][nz_start[i]...],
						// j are in sv_coef[i][nz_start[j]...]
						
						int si = start[i];
						int sj = start[j];
						int ci = count[i];
						int cj = count[j];
						
						int q = nz_start[i];
						int k;
						for (k = 0; k < ci; k++)
							if (nonzero[si + k])
								model.SupportVectorCoefficients[j - 1][q++] = f[p].alpha[k];
						q = nz_start[j];
						for (k = 0; k < cj; k++)
							if (nonzero[sj + k])
								model.SupportVectorCoefficients[i][q++] = f[p].alpha[ci + k];
						++p;
					}
			}
			return model;
		}
		
		public static void  svm_cross_validation(Problem prob, Parameter param, int nr_fold, double[] target)
		{
			int i;
			int[] perm = new int[prob.Count];
			
			// random shuffle
			for (i = 0; i < prob.Count; i++)
				perm[i] = i;
			for (i = 0; i < prob.Count; i++)
			{
				//UPGRADE_WARNING: Data types in Visual C# might be different.  Verify the accuracy of narrowing conversions. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1042_3"'
				int j = i + (int) (RAND.NextDouble() * (prob.Count - i));
				do 
				{
					int _ = perm[i]; perm[i] = perm[j]; perm[j] = _;
				}
				while (false);
			}
			for (i = 0; i < nr_fold; i++)
			{
				int begin = i * prob.Count / nr_fold;
				int end = (i + 1) * prob.Count / nr_fold;
				int j, k;
				Problem subprob = new Problem();
				
				subprob.Count = prob.Count - (end - begin);
				subprob.X = new Node[subprob.Count][];
				subprob.Y = new double[subprob.Count];
				
				k = 0;
				for (j = 0; j < begin; j++)
				{
					subprob.X[k] = prob.X[perm[j]];
					subprob.Y[k] = prob.Y[perm[j]];
					++k;
				}
				for (j = end; j < prob.Count; j++)
				{
					subprob.X[k] = prob.X[perm[j]];
					subprob.Y[k] = prob.Y[perm[j]];
					++k;
				}
				Model submodel = svm_train(subprob, param);
                if (param.Probability && (param.SvmType == SvmType.C_SVC || param.SvmType == SvmType.NU_SVC))
				{
					double[] prob_estimates = new double[svm_get_nr_class(submodel)];
					for (j = begin; j < end; j++)
						target[perm[j]] = svm_predict_probability(submodel, prob.X[perm[j]], prob_estimates);
				}
				else
					for (j = begin; j < end; j++)
						target[perm[j]] = svm_predict(submodel, prob.X[perm[j]]);
			}
		}

        public static SvmType svm_get_svm_type(Model model)
		{
			return model.Parameter.SvmType;
		}
		
		public static int svm_get_nr_class(Model model)
		{
			return model.NumberOfClasses;
		}
		
		public static void  svm_get_labels(Model model, int[] label)
		{
			if (model.ClassLabels != null)
				for (int i = 0; i < model.NumberOfClasses; i++)
					label[i] = model.ClassLabels[i];
		}
		
		public static double svm_get_svr_probability(Model model)
		{
            if ((model.Parameter.SvmType == SvmType.EPSILON_SVR || model.Parameter.SvmType == SvmType.NU_SVR) && model.PairwiseProbabilityA != null)
				return model.PairwiseProbabilityA[0];
			else
			{
				System.Console.Error.Write("Model doesn't contain information for SVR probability inference\n");
				return 0;
			}
		}
		
		public static void  svm_predict_values(Model model, Node[] x, double[] dec_values)
		{
            if (model.Parameter.SvmType == SvmType.ONE_CLASS || model.Parameter.SvmType == SvmType.EPSILON_SVR || model.Parameter.SvmType == SvmType.NU_SVR)
			{
				double[] sv_coef = model.SupportVectorCoefficients[0];
				double sum = 0;
				for (int i = 0; i < model.SupportVectorCount; i++)
					sum += sv_coef[i] * Kernel.k_function(x, model.SupportVectors[i], model.Parameter);
				sum -= model.Rho[0];
				dec_values[0] = sum;
			}
			else
			{
				int i;
				int nr_class = model.NumberOfClasses;
				int l = model.SupportVectorCount;
				
				double[] kvalue = new double[l];
				for (i = 0; i < l; i++)
					kvalue[i] = Kernel.k_function(x, model.SupportVectors[i], model.Parameter);
				
				int[] start = new int[nr_class];
				start[0] = 0;
				for (i = 1; i < nr_class; i++)
					start[i] = start[i - 1] + model.NumberOfSVPerClass[i - 1];
				
				int p = 0;
				int pos = 0;
				for (i = 0; i < nr_class; i++)
					for (int j = i + 1; j < nr_class; j++)
					{
						double sum = 0;
						int si = start[i];
						int sj = start[j];
						int ci = model.NumberOfSVPerClass[i];
						int cj = model.NumberOfSVPerClass[j];
						
						int k;
						double[] coef1 = model.SupportVectorCoefficients[j - 1];
						double[] coef2 = model.SupportVectorCoefficients[i];
						for (k = 0; k < ci; k++)
							sum += coef1[si + k] * kvalue[si + k];
						for (k = 0; k < cj; k++)
							sum += coef2[sj + k] * kvalue[sj + k];
						sum -= model.Rho[p++];
						dec_values[pos++] = sum;
					}
			}
		}
		
		public static double svm_predict(Model model, Node[] x)
		{
            if (model.Parameter.SvmType == SvmType.ONE_CLASS || model.Parameter.SvmType == SvmType.EPSILON_SVR || model.Parameter.SvmType == SvmType.NU_SVR)
			{
				double[] res = new double[1];
				svm_predict_values(model, x, res);

                if (model.Parameter.SvmType == SvmType.ONE_CLASS)
					return (res[0] > 0)?1:- 1;
				else
					return res[0];
			}
			else
			{
				int i;
				int nr_class = model.NumberOfClasses;
				double[] dec_values = new double[nr_class * (nr_class - 1) / 2];
				svm_predict_values(model, x, dec_values);
				
				int[] vote = new int[nr_class];
				for (i = 0; i < nr_class; i++)
					vote[i] = 0;
				int pos = 0;
				for (i = 0; i < nr_class; i++)
					for (int j = i + 1; j < nr_class; j++)
					{
						if (dec_values[pos++] > 0)
							++vote[i];
						else
							++vote[j];
					}
				
				int vote_max_idx = 0;
				for (i = 1; i < nr_class; i++)
					if (vote[i] > vote[vote_max_idx])
						vote_max_idx = i;
				return model.ClassLabels[vote_max_idx];
			}
		}
		
		public static double svm_predict_probability(Model model, Node[] x, double[] prob_estimates)
		{
            if ((model.Parameter.SvmType == SvmType.C_SVC || model.Parameter.SvmType == SvmType.NU_SVC) && model.PairwiseProbabilityA != null && model.PairwiseProbabilityB != null)
			{
				int i;
				int nr_class = model.NumberOfClasses;
				double[] dec_values = new double[nr_class * (nr_class - 1) / 2];
				svm_predict_values(model, x, dec_values);
				
				double min_prob = 1e-7;
				double[][] tmpArray = new double[nr_class][];
				for (int i2 = 0; i2 < nr_class; i2++)
				{
					tmpArray[i2] = new double[nr_class];
				}
				double[][] pairwise_prob = tmpArray;
				
				int k = 0;
				for (i = 0; i < nr_class; i++)
					for (int j = i + 1; j < nr_class; j++)
					{
						pairwise_prob[i][j] = System.Math.Min(System.Math.Max(sigmoid_predict(dec_values[k], model.PairwiseProbabilityA[k], model.PairwiseProbabilityB[k]), min_prob), 1 - min_prob);
						pairwise_prob[j][i] = 1 - pairwise_prob[i][j];
						k++;
					}
				multiclass_probability(nr_class, pairwise_prob, prob_estimates);
				
				int prob_max_idx = 0;
				for (i = 1; i < nr_class; i++)
					if (prob_estimates[i] > prob_estimates[prob_max_idx])
						prob_max_idx = i;
				return model.ClassLabels[prob_max_idx];
			}
			else
				return svm_predict(model, x);
		}
				
		public static void svm_check_parameter(Problem prob, Parameter param)
		{
			// svm_type

            SvmType svm_type = param.SvmType;
            if (svm_type != SvmType.C_SVC && svm_type != SvmType.NU_SVC && svm_type != SvmType.ONE_CLASS && svm_type != SvmType.EPSILON_SVR && svm_type != SvmType.NU_SVR)
				throw new Exception("unknown svm type");
			
			// kernel_type
			
			KernelType kernel_type = param.KernelType;
            if (kernel_type != KernelType.LINEAR && kernel_type != KernelType.POLY && kernel_type != KernelType.RBF && kernel_type != KernelType.SIGMOID)
                throw new Exception("unknown kernel type");
			
			// cache_size,eps,C,nu,p,shrinking
			
			if (param.CacheSize <= 0)
				throw new Exception("cache_size <= 0");
			
			if (param.EPS <= 0)
				throw new Exception("eps <= 0");

            if (svm_type == SvmType.C_SVC || svm_type == SvmType.EPSILON_SVR || svm_type == SvmType.NU_SVR)
				if (param.C <= 0)
					throw new Exception("C <= 0");

            if (svm_type == SvmType.NU_SVC || svm_type == SvmType.ONE_CLASS || svm_type == SvmType.NU_SVR)
				if (param.Nu < 0 || param.Nu > 1)
					throw new Exception("nu < 0 or nu > 1");

            if (svm_type == SvmType.EPSILON_SVR)
				if (param.P < 0)
					throw new Exception("p < 0");
			
            if (param.Probability && svm_type == SvmType.ONE_CLASS)
				throw new Exception("one-class SVM probability output not supported yet");
			
			// check whether nu-svc is feasible

            if (svm_type == SvmType.NU_SVC)
			{
				int l = prob.Count;
				int max_nr_class = 16;
				int nr_class = 0;
				int[] label = new int[max_nr_class];
				int[] count = new int[max_nr_class];
				
				int i;
				for (i = 0; i < l; i++)
				{
					//UPGRADE_WARNING: Data types in Visual C# might be different.  Verify the accuracy of narrowing conversions. 'ms-help://MS.VSCC.2003/commoner/redir/redirect.htm?keyword="jlca1042_3"'
					int this_label = (int) prob.Y[i];
					int j;
					for (j = 0; j < nr_class; j++)
						if (this_label == label[j])
						{
							++count[j];
							break;
						}
					
					if (j == nr_class)
					{
						if (nr_class == max_nr_class)
						{
							max_nr_class *= 2;
							int[] new_data = new int[max_nr_class];
							Array.Copy(label, 0, new_data, 0, label.Length);
							label = new_data;
							
							new_data = new int[max_nr_class];
							Array.Copy(count, 0, new_data, 0, count.Length);
							count = new_data;
						}
						label[nr_class] = this_label;
						count[nr_class] = 1;
						++nr_class;
					}
				}
				
				for (i = 0; i < nr_class; i++)
				{
					int n1 = count[i];
					for (int j = i + 1; j < nr_class; j++)
					{
						int n2 = count[j];
						if (param.Nu * (n1 + n2) / 2 > System.Math.Min(n1, n2))
							throw new Exception("specified nu is infeasible");
					}
				}
			}
		}
		
		public static int svm_check_probability_model(Model model)
		{
            if (((model.Parameter.SvmType == SvmType.C_SVC || model.Parameter.SvmType == SvmType.NU_SVC) && model.PairwiseProbabilityA != null && model.PairwiseProbabilityB != null) || ((model.Parameter.SvmType == SvmType.EPSILON_SVR || model.Parameter.SvmType == SvmType.NU_SVR) && model.PairwiseProbabilityA != null))
				return 1;
			else
				return 0;
		}
	}
}