//Copyright (C) 2007 Matthew Johnson

//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.

//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License along
//with this program; if not, write to the Free Software Foundation, Inc.,
//51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace SVM
{
    /// <summary>
    /// Class which encapsulates a range transformation.
    /// </summary>
    public class Range
    {
        private double[] _inputStart;
        private double[] _inputScale;
        private double _outputStart;
        private double _outputScale;
        private int _length;

        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="minValues">The minimum values in each dimension.</param>
        /// <param name="maxValues">The maximum values in each dimension.</param>
        /// <param name="lowerBound">The desired lower bound for all dimensions.</param>
        /// <param name="upperBound">The desired upper bound for all dimensions.</param>
        public Range(double[] minValues, double[] maxValues, double lowerBound, double upperBound)
        {
            _length = minValues.Length;
            if(maxValues.Length != _length)
                throw new Exception("Number of max and min values must be equal.");
            _inputStart = new double[_length];
            _inputScale = new double[_length];
            for (int i = 0; i < _length; i++)
            {
                _inputStart[i] = minValues[i];
                _inputScale[i] = maxValues[i] - minValues[i];
            }
            _outputStart = lowerBound;
            _outputScale = upperBound - lowerBound;
        }
        private Range(double[] inputStart, double[] inputScale, double outputStart, double outputScale, int length)
        {
            _inputStart = inputStart;
            _inputScale = inputScale;
            _outputStart = outputStart;
            _outputScale = outputScale;
            _length = length;
        }
        /// <summary>
        /// Transforms the input array based upon the values provided.
        /// </summary>
        /// <param name="input">The input array</param>
        /// <returns>A scaled array</returns>
        public double[] Transform(double[] input)
        {
            double[] output = new double[_length];
            for (int i = 0; i < _length; i++)
            {
                double tmp = input[i] - _inputStart[i];
                tmp /= _inputScale[i];
                tmp *= _outputScale;
                output[i] = tmp + _outputStart;
            }
            return output;
        }

        /// <summary>
        /// Transforms this an input value using the scaling transform for the provided dimension.
        /// </summary>
        /// <param name="input">The input value to transform</param>
        /// <param name="index">The dimension whose scaling transform should be used</param>
        /// <returns>The scaled value</returns>
        public double Transform(double input, int index)
        {
            double tmp = input - _inputStart[index];
            tmp /= _inputScale[index];
            tmp *= _outputScale;
            return tmp + _outputStart;
        }
        /// <summary>
        /// Writes this Range transform to a stream.
        /// </summary>
        /// <param name="stream">The stream to write to</param>
        /// <param name="r">The range to write</param>
        public static void Write(Stream stream, Range r)
        {
            BinaryWriter output = new BinaryWriter(stream);
            output.Write(r._length);
            for(int i=0; i<r._inputStart.Length; i++)
                output.Write(r._inputStart[i]);
            for (int i = 0; i < r._inputScale.Length; i++)
                output.Write(r._inputScale[i]);
            output.Write(r._outputStart);
            output.Write(r._outputScale);
            output.Flush();
        }

        /// <summary>
        /// Writes this Range transform to a file.    This will overwrite any previous data in the file.
        /// </summary>
        /// <param name="outputFile">The file to write to</param>
        /// <param name="r">The Range to write</param>
        public static void Write(string outputFile, Range r)
        {
            FileStream s = File.Open(outputFile, FileMode.Create);
            try
            {
                Write(s, r);
            }
            finally
            {
                s.Close();
            }
        }

        /// <summary>
        /// Reads a Range transform from a file.
        /// </summary>
        /// <param name="inputFile">The file to read from</param>
        /// <returns>The Range transform</returns>
        public static Range Read(string inputFile)
        {
            FileStream s = File.OpenRead(inputFile);
            try
            {
                return Read(s);
            }
            finally
            {
                s.Close();
            }
        }

        /// <summary>
        /// Reads a Range transform from a stream.
        /// </summary>
        /// <param name="stream">The stream to read from</param>
        /// <returns>The Range transform</returns>
        public static Range Read(Stream stream)
        {
            BinaryReader input = new BinaryReader(stream);
            int length = input.ReadInt32();
            double[] inputStart = new double[length];
            for (int i = 0; i < length; i++)
                inputStart[i] = input.ReadDouble();
            double[] inputScale = new double[length];
            for (int i = 0; i < length; i++)
                inputScale[i] = input.ReadDouble();
            double outputStart = input.ReadDouble();
            double outputScale = input.ReadDouble();
            return new Range(inputStart, inputScale, outputStart, outputScale, length);
        }
    }
}
